/*jslint devel: true, bitwise: true, regexp: true, browser: true, undef: true, unparam: true, white: true, newcap: true, nomen: true, plusplus: true, maxerr: 50, indent: 4 */
/*global $, Math, String, jQuery */
((function ($) {

	"use strict";
	
	$.widget("ui.tileslider", {
	
		options: {
			background:      null,
			horizontalTiles: 3,
			verticalTiles:   3,
			tileSize:        100,
			debug:           false,
			undone:          null,
			shuffled:        null,
			ready:           null
		},
		
		// model - must be reset later to create widget specific model
		grid: [[]],
		
		// history of moves made - must be reset later to create widget specific history
		history: [],
		
		// outer wrapper for tiles
		wrapper: null,
		
		// guid generation function
		_guid: function () {
			// http://stackoverflow.com/questions/105034/how-to-create-a-guid-uuid-in-javascript
			return "XXXXXXXX-XXXX-4XXX-YXXX-XXXXXXXXXXXX".replace(/[XY]/g, function (c) {
			    var r = Math.random() * 16 | 0, 
			        v = c === "X" ? r : (r & 0X3 | 0X8);
			    return v.toString(16);
			});
		},
		
		_tileDragStopHander: function (event, ui) {
			var self = {}, tile = {};
			self = event.data.owner;
			// get data
			tile = ui.helper.data("tile");
			// if tile has more than half of the tileSize then assume it has
			// been moved to a new position, otherwise revert back to the
			// original position
			if (Math.abs(ui.originalPosition.left - ui.position.left) > (self.options.tileSize / 2) || 
				Math.abs(ui.originalPosition.top  - ui.position.top)  > (self.options.tileSize / 2)) {
				// stick
				if (self.options.debug) {
					console.log("".concat("Moving tile ", String(tile.tile) , " from [", String(tile.move.from.y), ",", String(tile.move.from.x), "] to [", String(tile.move.to.y), ",", String(tile.move.to.x), "]"));
				}
				// complete move to new position
				ui.helper.animate({
					left: tile.move.to.x * self.options.tileSize,
					top:  tile.move.to.y * self.options.tileSize
				}, function () {
					// add to history
					self.history.push({
						from: {
							y: tile.move.from.y,
							x: tile.move.from.x
						},
						to: {
							y: tile.move.to.y,
							x: tile.move.to.x
						}
					});
					// swap positions
					self.grid[tile.move.to.y][tile.move.to.x]     = tile;
					self.grid[tile.move.from.y][tile.move.from.x] = null;
					// calculate new moves
					self._legalMoves();
					self._applyLegalMovesToDomGrid();
				});
				
			} else {
				// revert 
				if (self.options.debug) {
					console.log("".concat("Resetting tile to [", String(tile.move.from.y), ",", String(tile.move.from.x), "]"));
				}
				ui.helper.animate({
					left: tile.move.from.x * self.options.tileSize,
					top:  tile.move.from.y * self.options.tileSize
				});
			}
		},
		
		_legalMoves: function () {
			var w = 0, h = 0, wMax = 0, hMax = 0, tile = {};
			hMax = this.grid.length;
			for (h = 0; h < hMax; h++) {
				wMax = this.grid[h].length;
				for (w = 0; w < wMax; w++) {
					tile = this.grid[h][w];
					if (null !== tile) {
						// reset moves
						tile.moveUp    = false;
						tile.moveDown  = false;
						tile.moveLeft  = false;
						tile.moveRight = false;
						tile.move      = null;
						
						// is there an tile to..
						
						if (w > 0 && null === this.grid[h][w - 1]) {
							// the left
							tile.moveLeft  = true;	
							tile.move = {
								from: {
									y: h,
									x: w
								},
								to: {
									y: h,
									x: w - 1
								}
							};
						} else if (w + 1 < wMax && null === this.grid[h][w + 1]) {
							// or to the right
							tile.moveRight = true;
							tile.move = {
								from: {
									y: h,
									x: w
								},
								to: {
									y: h,
									x: w + 1
								}
							};
						} else if (h > 0 && null === this.grid[h - 1][w]) {
							// or above
							tile.moveUp = true;
							tile.move = {
								from: {
									y: h,
									x: w
								},
								to: {
									y: h - 1,
									x: w
								}
							};
						} else if (h + 1 < hMax && null === this.grid[h + 1][w]) {
							// or below
							tile.moveDown = true;
							tile.move = {
								from: {
									y: h,
									x: w
								},
								to: {
									y: h + 1,
									x: w
								}
							};
						}
					}
				}
			}
		},
		
		_initDomGrid: function () {
			var w, h, wMax, hMax, tile, domTile, self, inner;
			self = this;
			
			inner = $("<div class=\"inner\"><\/div>")
			.width(self.options.tileSize  * self.options.horizontalTiles)
			.height(self.options.tileSize * self.options.verticalTiles);
			
			hMax = this.grid.length;
			for (h = 0; h < hMax; h++) {
				wMax = this.grid[h].length;
				for (w = 0; w < wMax; w++) {
					tile = this.grid[h][w];
					if (null !== tile) {
						domTile = $("<div class=\"tile tile-" + String((w + wMax * h)) + "\"></div>")
						.width(self.options.tileSize)
						.height(self.options.tileSize)
						.css({
							"top":                 h * this.options.tileSize,
							"left":                w * this.options.tileSize,
							"background-image":    "".concat("url(\"", this.element.attr("src") ,"\")"),
							"background-position": "-" + String(w * this.options.tileSize) + "px -" + String(h * this.options.tileSize) + "px"
						})
						.draggable()
						.data({
							tile: tile
						});
						
						/*
						 * instead of using the code below and creating a function
						 * in a loop, bind to the dragstart event instead
						.draggable({
							stop: function (event, ui) {
								self._tileDragStopHander(ui, self);
							}
						})
						*/
						domTile.bind("dragstop", {owner: self}, self._tileDragStopHander);
						
						inner.append(domTile);
					}
				}
			}
			this.wrapper.append(inner);
			
			// add event handler to prevent dragging of images (Firefox issue)
			$("div.tileslider img, div.tile").mousedown(function (e) {
				e.preventDefault();
			});
		},
		
		_initGrid: function () {
			var w, h;
            // create widget specific model and history
			this.grid = [];
            this.history = [];
			// size grid and set default positions
			for (h = 0; h < this.options.verticalTiles; h++) {
				this.grid[h] = [];
				for (w = 0; w < this.options.horizontalTiles; w++) {
					this.grid[h][w] = {
						tile:      h * this.options.horizontalTiles + w,
						moveUp:    false,
						moveDown:  false,
						moveLeft:  false,
						moveRight: false
					};
				}
			}
			// blank last tile
			this.grid[this.options.verticalTiles - 1][this.options.horizontalTiles - 1] = null;
		},
	
		_applyLegalMovesToDomGrid: function () {
			var w, h, wMax, hMax, tile, domTile, slider, axis, sliderCounter, sliderId, sliderTop, sliderLeft, sliderWidth, sliderHeight;
			
			sliderCounter = 0;
			
			// remove existing sliders
			this.wrapper.find(".slider").remove();
			
			hMax = this.grid.length;
			for (h = 0; h < hMax; h++) {
				wMax = this.grid[h].length;
				for (w = 0; w < wMax; w++) {
					tile = this.grid[h][w];
					if (null !== tile) {
						
						// reset values
						domTile    = this.wrapper.find("".concat(".tile-", String(tile.tile))); 
						axis       = null;
						sliderTop  = -1;
						sliderLeft = -1;
						
						if (tile.moveUp) {
							axis         = "y";
							sliderTop    = domTile.position().top - this.options.tileSize;
							sliderLeft   = domTile.position().left;
							sliderWidth  = this.options.tileSize;
							sliderHeight = 2 * this.options.tileSize;
						} else if (tile.moveDown) {
							axis         = "y";
							sliderTop    = domTile.position().top;
							sliderLeft   = domTile.position().left;
							sliderWidth  = this.options.tileSize;
							sliderHeight = 2 * this.options.tileSize;
						} else if (tile.moveLeft) {
							axis         = "x";
							sliderTop    = domTile.position().top;
							sliderLeft   = domTile.position().left - this.options.tileSize;
							sliderWidth  = 2 * this.options.tileSize;
							sliderHeight = this.options.tileSize;
						} else if (tile.moveRight) {
							axis         = "x";
							sliderTop    = domTile.position().top;
							sliderLeft   = domTile.position().left;
							sliderWidth  = 2 * this.options.tileSize;
							sliderHeight = this.options.tileSize;
						}
						
						if (axis && sliderTop >= 0 && sliderLeft >= 0) {
							// tile may slide
							sliderId = this._guid();
							sliderCounter++;
							slider = $("".concat("<div id=\"", sliderId, "\" class=\"slider\"><\/div>"))
							.css({
								top:  "".concat(String(sliderTop),  "px"),
								left: "".concat(String(sliderLeft), "px")
							})
							.width(sliderWidth)
							.height(sliderHeight);

							this.wrapper.append(slider);

							domTile.draggable("option", {
								"axis":        axis,
								"containment": "".concat("#", sliderId),
								"disabled":    false,
								"snap":        "inner"
							});
						} else {
							// tile is locked
							domTile.draggable("option", "disabled", true);
						}
					}
				}
			}
		},
		
		_moveableTiles: function (exclude) {
			var w, h, wMax, hMax, tile, moveable;
			moveable = [];			
			hMax = this.grid.length;
			for (h = 0; h < hMax; h++) {
				wMax = this.grid[h].length;
				for (w = 0; w < wMax; w++) {
					tile = this.grid[h][w];
					if (null !== tile && null !== tile.move) {
						if (exclude.tile   !== tile.tile || 
							(exclude.tile   === tile.tile &&
						     exclude.from.x !== tile.move.to.x &&
						     exclude.from.y !== tile.move.to.y)) {
							moveable.push({
								x: w,
								y: h
							});
						}
					}
				}
			}
			return moveable;
		},
		
		_shuffler: function (data) {
			var moveable, r, tile, domTile, self;
			if (data.moves <= 0) {
				this._applyLegalMovesToDomGrid();
				this.element.trigger("tileslidershuffled");
				
			} else {
				self = this;
				// pick moveable piece
				moveable = this._moveableTiles(data);
				r = Math.floor(Math.random() * moveable.length);
				tile = self.grid[moveable[r].y][moveable[r].x];
				// move piece
				domTile = self.wrapper.find("".concat(".tile-", String(tile.tile))); 
				domTile.animate({
					left: tile.move.to.x * self.options.tileSize,
					top:  tile.move.to.y * self.options.tileSize
				}, function () {
					// add to history
					self.history.push({
						from: {
							y: tile.move.from.y,
							x: tile.move.from.x
						},
						to: {
							y: tile.move.to.y,
							x: tile.move.to.x
						}
					});
					// swap positions
					self.grid[tile.move.to.y][tile.move.to.x] = tile;
					self.grid[tile.move.from.y][tile.move.from.x] = null;
					// calculate new moves
					self._legalMoves();
					// next move
					self._shuffler({
						moves: data.moves - 1, 
						tile:  tile.tile,
						from:  {
							y: tile.move.to.y,
							x: tile.move.to.x
						}
					});
				});
			}
		},
		
		shuffle: function (moves) {
			// disable dragging
			this.wrapper.find(".tile").draggable("option", "disabled", true);
			this._shuffler({
				moves: moves, 
				tile:  -1,
				from:  {
					y: -1, 
					x: -1
				}
			});
		},
		
		_undoer: function () {
			var move, tile, domTile, self;
			self = this;
			if (0 === self.history.length) {
				// calculate new moves
				self._legalMoves();
				self._applyLegalMovesToDomGrid();
				this.element.trigger("tilesliderundone");
				
			} else {
				move = self.history.pop();
				tile = self.grid[move.to.y][move.to.x];
				// move piece
				domTile= self.wrapper.find("".concat(".tile-", String(tile.tile))); 
				domTile.animate({
					left: move.from.x * self.options.tileSize,
					top:  move.from.y * self.options.tileSize
				}, function () {
					// swap positions
					self.grid[move.to.y][move.to.x] = null;
					self.grid[move.from.y][move.from.x] = tile;
					// next undo
					self._undoer();
				});
			}
		},
		
		undo: function () {
			// disable dragging
			this.wrapper.find(".tile").draggable("option", "disabled", true);
			this._undoer();
		},
		
		model: function () {
			return this.grid;
		},
		
        movehistory: function () {
            return this.history;
        },
        
		// event handlers 
		_configureChangeEvent: function (eventname, handler) {
			this.element.unbind("".concat("tileslider", eventname));
			if ("function" === typeof handler) {
				this.options[eventname] = handler;
				this.element.bind("".concat("tileslider", eventname), function (e) {
					handler(e);
				});
			} else {
				this.options[eventname] = null;
			}
		},
		
		// called once on creation
		_create: function () {
			// check element is image
			
			// configure DOM
			this.wrapper = this.element
			// wrap in div
			.wrap("<div class=\"tileslider\"><\/div>")	
			// add other items as siblings to original input
			.parent()
			// size wrapper
			.width(this.options.horizontalTiles * this.options.tileSize)
			.height(this.options.verticalTiles * this.options.tileSize);
			
			// hide original image
			this.element.hide();
			
			// event handlers
			this._configureChangeEvent("ready",    this.options.ready);
			this._configureChangeEvent("shuffled", this.options.shuffled);
			this._configureChangeEvent("undone",   this.options.undone);
			
			// use background?
			if (this.options.background) {
				this.wrapper.append("".concat("<img src=\"", this.options.background, "\" class=\"background\"\/>"));
			}
			
		},
		
		// called everytime .tileslider() is called with no parameters
		_init: function () {
			// create model of grid
			this._initGrid();
			// draw grid in DOM
			this._initDomGrid();
			// configure legal moves
			this._legalMoves();
			// enable legal moves
			this._applyLegalMovesToDomGrid(true);
			// ready
			this.element.trigger("tilesliderready");
		},

		_setOption: function (key, value) {
			if ("ready" === key || "shuffled" === key || "undone" === key) {	
				this._configureChangeEvent(key, value);
			}
            $.Widget.prototype._setOption.apply(this, arguments);
		},
		
		destroy: function () {
			$.Widget.prototype.destroy.apply(this, arguments); // default destroy
			// now do other stuff particular to this widget
			// remove everything in wrapper apart from image
			this.wrapper.children().not("img").remove();
			this.element.show().unwrap();
		}
	});
	
}(jQuery)));