/*jslint white: true, onevar: true, undef: true, nomen: true, regexp: true, plusplus: true, bitwise: true, newcap: true, browser: true, devel: true, maxerr: 50, indent: 4 */
/*global $, google, window, expect, equal, module, start, stop, test, MOCKUP, MAP */
(function () {
    "use strict";

	module("TILES Tests", {
		setup: function () {
		},
		teardown: function () {
		}
	});

	/**
	 * Create with Defaults.
	 */
	test('Create with Defaults', 2, function () {
		
		var proxyTiles;
		
		proxyTiles = $("#tileslider").tileslider();
		
		equal(proxyTiles.tileslider("option", "horizontalTiles"),  3, "Default number of horizontal tiles is 3");
		equal(proxyTiles.tileslider("option", "verticalTiles"), 3, "Default number of vertical tiles is 3");
		
	});

	/**
	 * Create with Non Defaults.
	 */
	test('Create with Non Defaults', 2, function () {
		
		var proxyTiles;
		
		proxyTiles = $("#tileslider").tileslider({
			horizontalTiles: 5,
			verticalTiles: 5
		});
		
		equal(proxyTiles.tileslider("option", "horizontalTiles"),  5, "Non default number of horizontal tiles is 5");
		equal(proxyTiles.tileslider("option", "verticalTiles"), 5,  "Non default number of vertical tiles is 5");
		
	});
	 
	/**
	 * Create Grid with Defaults
	 */
	test('Init Grid with Defaults', 13, function () {
		
		var proxyTiles, model;
		
		proxyTiles = $("#tileslider").tileslider();
		
		// initial settings
		equal(proxyTiles.tileslider("option", "horizontalTiles"),  3, "Default number of horizontal tiles is 3");
		equal(proxyTiles.tileslider("option", "verticalTiles"), 3, "Default number of vertical tiles is 3");
		
		model = proxyTiles.tileslider("model");
		
		// initial grid sizes
		equal(model.length,    3, "Default grid width is 3");
		equal(model[0].length, 3, "Default grid width is 3");
		
		// initial tile positions and moves
		deepEqual(model[0][0], {tile: 0, moveUp: false, moveDown: false, moveLeft: false, moveRight: false, move: null}, "Value in grid at [0,0] is 0, no moves");
		deepEqual(model[0][1], {tile: 1, moveUp: false, moveDown: false, moveLeft: false, moveRight: false, move: null}, "Value in grid at [0,1] is 1, no moves");
		deepEqual(model[0][2], {tile: 2, moveUp: false, moveDown: false, moveLeft: false, moveRight: false, move: null}, "Value in grid at [0,2] is 2, no moves");

		deepEqual(model[1][0], {tile: 3, moveUp: false, moveDown: false, moveLeft: false, moveRight: false, move: null}, "Value in grid at [1,0] is 3, no moves");
		deepEqual(model[1][1], {tile: 4, moveUp: false, moveDown: false, moveLeft: false, moveRight: false, move: null}, "Value in grid at [1,1] is 4, no moves");
		deepEqual(model[1][2], {tile: 5, moveUp: false, moveDown: true,  moveLeft: false, moveRight: false, move: {from: {x: 2, y: 1}, to: {x: 2, y: 2}}}, "Value in grid at [1,2] is 5, can move down");

		deepEqual(model[2][0], {tile: 6, moveUp: false, moveDown: false, moveLeft: false, moveRight: false, move: null}, "Value in grid at [2,0] is 6, no moves");
		deepEqual(model[2][1], {tile: 7, moveUp: false, moveDown: false, moveLeft: false, moveRight: true, move: {from: {x: 1, y: 2}, to: {x: 2, y: 2}}}, "Value in grid at [2,1] is 7, can move right");
		deepEqual(model[2][2], null, "Value in grid at [2,2] is null");
	});
	
}());